----------------------------------------------------------------------------
--	GPIO_Demo.vhd -- Nexys Video GPIO/UART Demonstration Project
----------------------------------------------------------------------------
-- Author:  Samuel Lowe adapted from..
--          Marshall Wingerson Adapted from Sam Bobrowicz
--          Copyright 2015 Digilent, Inc.
----------------------------------------------------------------------------
-- Modified for MQP by Andrew Gray
----------------------------------------------------------------------------
--	The GPIO/UART Demo project demonstrates a simple usage of the Nexys4DDR's 
--  GPIO and UART. The behavior is as follows:
--
--	      *The 8 User LEDs are tied to the 8 User Switches.
--       *An introduction message is sent across the UART when the device
--        is finished being configured, and after the center User button
--        is pressed.
--       *A message is sent over UART whenever BTNU, BTNL, BTND, or BTNR is
--        pressed.
--        
--	All UART communication can be captured by attaching the UART port to a
-- computer running a Terminal program with 9600 Baud Rate, 8 data bits, no 
-- parity, and 1 stop bit.																
----------------------------------------------------------------------------
--
----------------------------------------------------------------------------
-- Revision History:
--  08/08/2011(SamB): Created using Xilinx Tools 13.2
--  08/27/2013(MarshallW): Modified for the Nexys4 with Xilinx ISE 14.4\
--  		--added RGB and microphone
--  12/10/2014(SamB): Ported to Nexys4DDR and updated to Vivado 2014.4
--  12/10/2014(SamL): Ported to Nexys Video and updated to Vivado 2015.1
----------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

--The IEEE.std_logic_unsigned contains definitions that allow 
--std_logic_vector types to be used with the + operator to instantiate a 
--counter.
use IEEE.std_logic_unsigned.all;

entity UART_CALL is
generic (LEN: integer :=10;
NUM_EQs: integer:=6; TERMS: integer:=4; LEN_TERMS:integer:=9);
    Port ( 
           TERMS_V                 : in   STD_LOGIC_VECTOR (LEN_TERMS downto 0);
           BTN 			        : in   STD_LOGIC_VECTOR (4 downto 0);
           EQ                   : in STD_LOGIC_VECTOR ((NUM_EQs*LEN)-1 downto 0); 
           CLK 			        : in   STD_LOGIC;           
           UART_TXD 	        : out  STD_LOGIC       
          );
end UART_CALL;

architecture Behavioral of UART_CALL is

component UART_TX_CTRL
Port(
	SEND : in std_logic;
	DATA : in std_logic_vector(7 downto 0);
	CLK : in std_logic;
	READY : out std_logic;
	UART_TX : out std_logic
	);
end component;

--The type definition for the UART state machine type. Here is a description of what
--occurs during each state:
-- RST_REG     -- Do Nothing. This state is entered after configuration or a user reset.
--                The state is set to LD_INIT_STR.
-- LD_INIT_STR -- The Welcome String is loaded into the sendStr variable and the strIndex
--                variable is set to zero. The welcome string length is stored in the StrEnd
--                variable. The state is set to SEND_CHAR.
-- SEND_CHAR   -- uartSend is set high for a single clock cycle, signaling the character
--                data at sendStr(strIndex) to be registered by the UART_TX_CTRL at the next
--                cycle. Also, strIndex is incremented (behaves as if it were post 
--                incremented after reading the sendStr data). The state is set to RDY_LOW.
-- RDY_LOW     -- Do nothing. Wait for the READY signal from the UART_TX_CTRL to go low, 
--                indicating a send operation has begun. State is set to WAIT_RDY.
-- WAIT_RDY    -- Do nothing. Wait for the READY signal from the UART_TX_CTRL to go high, 
--                indicating a send operation has finished. If READY is high and strEnd = 
--                StrIndex then state is set to WAIT_BTN, else if READY is high and strEnd /=
--                StrIndex then state is set to SEND_CHAR.
-- WAIT_BTN    -- Do nothing. Wait for a button press on BTNU, BTNL, BTND, or BTNR. If a 
--                button press is detected, set the state to LD_BTN_STR.
-- LD_BTN_STR  -- The Button String is loaded into the sendStr variable and the strIndex
--                variable is set to zero. The button string length is stored in the StrEnd
--                variable. The state is set to SEND_CHAR.
type UART_STATE_TYPE is (RST_REG, LD_INIT_STR, SEND_CHAR, RDY_LOW, WAIT_RDY, WAIT_BTN, LD_BTN_STR);

--The CHAR_ARRAY type is a variable length array of 8 bit std_logic_vectors. 
--Each std_logic_vector contains an ASCII value and represents a character in
--a string. The character at index 0 is meant to represent the first
--character of the string, the character at index 1 is meant to represent the
--second character of the string, and so on.
type CHAR_ARRAY is array (integer range<>) of std_logic_vector(7 downto 0);

constant TMR_CNTR_MAX : std_logic_vector(26 downto 0) := "101111101011110000100000000"; --100,000,000 = clk cycles per second
constant TMR_VAL_MAX : std_logic_vector(3 downto 0) := "1001"; --9

constant RESET_CNTR_MAX : std_logic_vector(17 downto 0) := "110000110101000111";-- 100,000,000 * 0.002 = 200,000 = clk cycles per 2 ms

constant MAX_STR_LEN : integer := 200;
constant WELCOME_STR_LEN : natural := 31;
constant BTN_STR_LEN : natural := 25;

--Welcome string definition. Note that the values stored at each index
--are the ASCII values of the indicated character.
constant WELCOME_STR : CHAR_ARRAY(0 to 30) := 				 (X"0A",  --\n
															  X"0D",  --\r
															  X"45",  --e
															  X"43",  --c
															  X"45",  --e
															  X"20",  
															  X"4D",   --m
															  X"51",  --q
															  X"50",  --p
															  X"20",  
															  X"55",  --U
															  X"41",  --A
															  X"52",  --R 
															  X"54",  --T
															  X"20",  --
															  X"46",  --f
															  X"50",  --p
															  X"47",  --g
															  X"41",  --a 
															 X"20",  
															  X"32",  --2
															  X"30",  --0
															  X"32",  --2
															  X"33",  --33
															  X"2F",  --/
															  X"32",  --2
															  X"34",  --4
															  X"20",  --E
															  X"20",  --M
															  X"20",  --O 
															 
															  X"0D"); --\r													  
--Button press string definition.
  -------------------------------------------------------------------
    --                        ASCII HEX TABLE
    --  Hex                 Low Hex Digit
    -- Value  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
    ------\----------------------------------------------------------------
    --H  2 |  SP  !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /
    --i  3 |  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
    --g  4 |  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
    --h  5 |  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _
    --   6 |  `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
    --   7 |  p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~ DEL
    -----------------------------------------------------------------------
constant BTN_STR : CHAR_ARRAY(0 to BTN_STR_LEN) :=     				 (X"0A",X"23", X"20",
X"58", X"31",X"20",
 X"58", X"32", X"20",
  X"58",X"33", X"20",
  X"58", X"31", X"33", X"20",
    X"58", X"32", X"33", X"20",
     X"58", X"33", X"32", X"20", X"0A", X"0A" ); 

--This is used to determine when the 7-segment display should be
--incremented
signal tmrCntr : std_logic_vector(26 downto 0) := (others => '0');

--Contains the current string being sent over uart.
signal sendStr : CHAR_ARRAY(0 to (MAX_STR_LEN - 1));

--Contains the length of the current string being sent over uart.
signal strEnd : natural;

--Contains the index of the next character to be sent over uart
--within the sendStr variable.
signal strIndex : natural;

--Used to determine when a button press has occured
signal btnReg : std_logic_vector (4 downto 0) := "00000";
signal btnDetect : std_logic;

--UART_TX_CTRL control signals
signal uartRdy : std_logic;
signal uartSend : std_logic := '0';
signal uartData : std_logic_vector (7 downto 0):= "00000000";
signal uartTX : std_logic;

--Current uart state signal
signal uartState : UART_STATE_TYPE := RST_REG;

--Debounced btn signals used to prevent single button presses
--from being interpreted as multiple button presses.
signal btnDeBnc : std_logic_vector(4 downto 0);

signal clk_cntr_reg : std_logic_vector (4 downto 0) := (others=>'0'); 

--signal pwm_val_reg : std_logic := '0';

--this counter counts the amount of time paused in the UART reset state
signal reset_cntr : std_logic_vector (17 downto 0) := (others=>'0');


--LEN
 -- constant total:integer:LEN+2;
constant BTN_EQ_LEN: natural:=(((LEN)*NUM_EQs)+NUM_EQs+LEN_TERMS+1);
constant EQ_LEN: natural:=(((LEN)*NUM_EQs)-1);
constant EQ_LEN2:natural:=EQ_LEN+1;
signal BTN_EQ_print: CHAR_ARRAY(0 to (BTN_EQ_LEN));
signal BTN_EQ_flip: std_logic_vector(LEN downto (0));

--signal BTN_EQ_flip:std_logic_vector (LEN+1 downto 0);
signal x:integer:=0;
signal y:integer:=0;
signal xt:integer:=0;
signal start:std_logic:='0';
--signal EQ_0: std_logic_vector;
begin

--BTN_EQ_print (0)<= X"0A"; --#
looper : process  begin 

   for x in 0 to (EQ_LEN) loop       
     BTN_EQ_print (x)<= X"31" when (EQ(x)='1') else X"30" after 10ns;
 end loop;
 --for y in 1 to NUM_EQs
 for y in 1 to NUM_EQs loop
 BTN_EQ_print((y*LEN)-1)<=X"20";

 end loop;

BTN_EQ_print((EQ_LEN))<=X"20";
 
 for y in LEN_TERMS downto 1 loop
 if (y mod (TERMS+1)=0) then
 BTN_EQ_print(EQ_LEN+y+1)<=X"20";
 else

 BTN_EQ_print (EQ_LEN2+y)<= X"31" when (TERMS_V(y)='1') else X"30" after 10ns;
end if;
 end loop;


end process;

btnDeBnc<=BTN;

  

----------------------------------------------------------
------                LED Control                  -------
----------------------------------------------------------

-- Tie the LEDs straight to the switches

	--LED <= SW;
			 			 

----------------------------------------------------------
------              Button Control                 -------
----------------------------------------------------------
--Buttons are debounced and their rising edges are detected
--to trigger UART messages




--Registers the debounced button signals, for edge detection.
btn_reg_process : process (CLK)
begin
	if (rising_edge(CLK)) then
		btnReg <= btnDeBnc(4 downto 0);
	end if;
end process;

--btnDetect goes high for a single clock cycle when a btn press is
--detected. This triggers a UART message to begin being sent.
btnDetect <= '1' when ((btnReg(0)='0' and btnDeBnc(0)='1') or
								(btnReg(1)='0' and btnDeBnc(1)='1') or
								(btnReg(2)='0' and btnDeBnc(2)='1') or
								(btnReg(3)='0' and btnDeBnc(3)='1') or
								(btnReg(4)='0' and btnDeBnc(4)='1') ) else
				  '0';
				  



----------------------------------------------------------
------              UART Control                   -------
----------------------------------------------------------
--Messages are sent on reset and when a button is pressed.

--This counter holds the UART state machine in reset for ~2 milliseconds. This
--will complete transmission of any byte that may have been initiated during 
--FPGA configuration due to the UART_TX line being pulled low, preventing a 
--frame shift error from occuring during the first message.
process(CLK)
begin
  if (rising_edge(CLK)) then
    if ((reset_cntr = RESET_CNTR_MAX) or (uartState /= RST_REG)) then
      reset_cntr <= (others=>'0');
    else
      reset_cntr <= reset_cntr + 1;
    end if;
  end if;
end process;

--Next Uart state logic (states described above)
next_uartState_process : process (CLK)
begin
	if (rising_edge(CLK)) then
			
			case uartState is 
			when RST_REG =>
        if (reset_cntr = RESET_CNTR_MAX) then
          uartState <= LD_INIT_STR;
        end if;
			when LD_INIT_STR =>
				uartState <= SEND_CHAR;
			when SEND_CHAR =>
				uartState <= RDY_LOW;
			when RDY_LOW =>
				uartState <= WAIT_RDY;
			when WAIT_RDY =>
				if (uartRdy = '1') then
					if (strEnd = strIndex) then
						uartState <= WAIT_BTN;
					else
						uartState <= SEND_CHAR;
					end if;
				end if;
			when WAIT_BTN =>
				if (btnDetect = '1') then
					uartState <= LD_BTN_STR;
				end if;
			when LD_BTN_STR =>
				uartState <= SEND_CHAR;
			when others=> --should never be reached
				uartState <= RST_REG;
			end case;
		
	end if;
end process;

--Loads the sendStr and strEnd signals when a LD state is
--is reached.
string_load_process : process (CLK)
begin
	if (rising_edge(CLK)) then
		if (uartState = LD_INIT_STR) then
			sendStr(0 to BTN_EQ_LEN) <= BTN_EQ_print; --welcome str
			strEnd <= BTN_EQ_LEN; --welcome len
		elsif (uartState = LD_BTN_STR) then
			sendStr(0 to BTN_STR_LEN) <= BTN_STR; --btn eq print
			strEnd <= BTN_STR_LEN; --btn eq len
		end if;
	end if;
end process;

--Conrols the strIndex signal so that it contains the index
--of the next character that needs to be sent over uart
char_count_process : process (CLK)
begin
	if (rising_edge(CLK)) then
		if (uartState = LD_INIT_STR or uartState = LD_BTN_STR) then
			strIndex <= 0;
		elsif (uartState = SEND_CHAR) then
			strIndex <= strIndex + 1;
		end if;
	end if;
end process;

--Controls the UART_TX_CTRL signals
char_load_process : process (CLK)
begin
	if (rising_edge(CLK)) then
		if (uartState = SEND_CHAR) then
			uartSend <= '1';
			uartData <= sendStr(strIndex);
		else
			uartSend <= '0';
		end if;
	end if;
end process;

--Component used to send a byte of data over a UART line.
Inst_UART_TX_CTRL: UART_TX_CTRL port map( 
		SEND => uartSend,
		DATA => uartData,
		CLK => CLK,
		READY => uartRdy,
		UART_TX => uartTX 
	);

UART_TXD <= uartTX;
end Behavioral;